import os
import time
import unittest
import xml.etree.ElementTree as ET
from math import nan

import pytest

from asyncfix import FMsg, FTag
from asyncfix.errors import FIXError
from asyncfix.protocol.common import FExecType, FOrdSide, FOrdStatus, FOrdType
from asyncfix.protocol.fix_tester import FIXTester
from asyncfix.protocol.order_single import FIXNewOrderSingle
from asyncfix.protocol.schema import FIXSchema

TEST_DIR = os.path.abspath(os.path.dirname(__file__))
fix44_schema = ET.parse(os.path.join(TEST_DIR, "FIX44.xml"))
FIX_SCHEMA = FIXSchema(fix44_schema)


def test_init_order_single_default_short():
    ord_dict = {}
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.SELL, price=100.0, qty=20
    )
    assert isinstance(o, FIXNewOrderSingle)
    assert o.status == FOrdStatus.CREATED
    assert o.ord_type == FOrdType.LIMIT
    assert o.price == 100
    assert o.qty == 20
    assert o.leaves_qty == 0
    assert o.cum_qty == 0
    assert o.clord_id == "clordTest"
    assert o.orig_clord_id is None
    assert o.side == FOrdSide.SELL
    assert o.target_price == 100
    assert o.ord_type == FOrdType.LIMIT

    m = o.new_req()

    # Account
    assert m[FTag.Account] == "000000"

    # Tag 11: ClOrdID
    assert m[FTag.ClOrdID] == "clordTest"

    # Tag 38: Order Qty
    assert m[FTag.OrderQty] == "20"

    # Tag 40: Order Type
    assert m[FTag.OrdType] == "2"  # Limit order

    # Tag 44: Order Price
    assert m[FTag.Price] == "100.0"

    # Tag 54: Side
    assert m[54] == "2"  # sell

    #  Tag 55: Symbol set to v2 symbol
    assert m[FTag.Symbol] == "US.F.TICKER"

    # Overall message is valid!
    assert FIX_SCHEMA.validate(m)


def test_init_order_single_long():
    o = FIXNewOrderSingle(
        "clordTest",
        "US.F.TICKER",
        side=FOrdSide.BUY,
        price=200.0,
        qty=20,
        target_price=220,
        ord_type=FOrdType.MARKET,
    )
    assert isinstance(o, FIXNewOrderSingle)
    assert o.status == FOrdStatus.CREATED
    assert o.price == 200
    assert o.qty == 20
    assert o.leaves_qty == 0
    assert o.cum_qty == 0
    assert o.clord_id == "clordTest"
    assert o.orig_clord_id is None
    assert o.side == FOrdSide.BUY
    assert o.target_price == 220
    assert o.ord_type == FOrdType.MARKET

    m = o.new_req()

    # Account
    assert m[1] == "000000"

    # Tag 11: ClOrdID
    assert m[11] == "clordTest"

    # Tag 38: Order Qty
    assert m[38] == "20"

    # Tag 40: Order Type
    assert m[40] == "1"

    # Tag 44: Order Price
    assert m[44] == "200.0"

    # Tag 54: Side
    assert m[54] == "1"  # buy

    #  Tag 55: Symbol set to v2 symbol
    assert m[55] == "US.F.TICKER"

    # Overall message is valid!
    assert FIX_SCHEMA.validate(m)


def test_simple_execution_report_state_created__2__pending_new():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=20
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PENDING_NEW, f"o.status={chr(o.status)}"


def test_simple_execution_report_state_created__2__rejected():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=20
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(o, o.clord_id, FExecType.REJECTED, FOrdStatus.REJECTED)
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.REJECTED, f"o.status={chr(o.status)}"


def test_state_transition__unsupported_msg():
    # fmt: off
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, FMsg.ADVERTISEMENT, 0, FOrdStatus.PENDING_NEW) 
    # fmt: on


def test_state_transition__created__execution_report():
    # fmt: off
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.CREATED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.NEW) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.FILLED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.CANCELED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.STOPPED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.REJECTED) == FOrdStatus.REJECTED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) == FOrdStatus.PENDING_NEW
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.CALCULATED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.EXPIRED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.CREATED, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) 
    # fmt: on


def test_state_transition__pendingnew__execution_report():
    # fmt: off
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.CREATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.NEW) == FOrdStatus.NEW
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) == FOrdStatus.PARTIALLY_FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.FILLED) == FOrdStatus.FILLED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.CANCELED) == FOrdStatus.CANCELED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.STOPPED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.REJECTED) == FOrdStatus.REJECTED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) == FOrdStatus.SUSPENDED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.CALCULATED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.EXPIRED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_NEW, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) 
    # fmt: on


def test_state_transition__new__execution_report():
    # fmt: off
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.CREATED)
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.CREATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) == FOrdStatus.PARTIALLY_FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.FILLED) == FOrdStatus.FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) == FOrdStatus.DONE_FOR_DAY
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.CANCELED) == FOrdStatus.CANCELED
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) == FOrdStatus.PENDING_CANCEL
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.STOPPED) == FOrdStatus.STOPPED
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.REJECTED) == FOrdStatus.REJECTED
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) == FOrdStatus.SUSPENDED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.CALCULATED) == FOrdStatus.CALCULATED
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.EXPIRED) == FOrdStatus.EXPIRED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.NEW, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) == FOrdStatus.PENDING_REPLACE
    # fmt: on


def test_state_transition__rejected__execution_report():
    # fmt: off
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.CREATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.CANCELED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.STOPPED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.REJECTED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.CALCULATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.EXPIRED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.REJECTED, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) is None
    # fmt: on


def test_state_transition__filled__execution_report():
    # fmt: off
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.CREATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.CANCELED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.STOPPED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.REJECTED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.CALCULATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.EXPIRED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.FILLED, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) is None
    # fmt: on


def test_state_transition__expired__execution_report():
    # fmt: off
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.CREATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.CANCELED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.STOPPED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.REJECTED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.CALCULATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.EXPIRED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.EXPIRED, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) is None
    # fmt: on


def test_state_transition__canceled__execution_report():
    # fmt: off
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.CREATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.CANCELED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.STOPPED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.REJECTED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.CALCULATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.EXPIRED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.CANCELED, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) is None
    # fmt: on


def test_state_transition__suspended__execution_report():
    # fmt: off
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.CREATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.NEW) == FOrdStatus.NEW
    assert FIXNewOrderSingle.change_status(FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) == FOrdStatus.PARTIALLY_FILLED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.FILLED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.CANCELED) == FOrdStatus.CANCELED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.STOPPED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.REJECTED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) is None
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.CALCULATED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.EXPIRED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.SUSPENDED, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) 
    # fmt: on


def test_state_transition__partiallyfilled__execution_report():
    # fmt: off
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.CREATED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.NEW) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) == FOrdStatus.PARTIALLY_FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.FILLED) == FOrdStatus.FILLED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.CANCELED) == FOrdStatus.CANCELED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) == FOrdStatus.PENDING_CANCEL
    assert FIXNewOrderSingle.change_status(FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.STOPPED) == FOrdStatus.STOPPED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.REJECTED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) == FOrdStatus.SUSPENDED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.CALCULATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.EXPIRED) == FOrdStatus.EXPIRED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PARTIALLY_FILLED, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) == FOrdStatus.PENDING_REPLACE
    # fmt: on


def test_state_transition__pendingcancel__execution_report():
    # fmt: off
    # Executin report doesn't not have any effect of pending cancelled state
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.CREATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.CANCELED) == FOrdStatus.CANCELED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.STOPPED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.REJECTED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.CALCULATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.EXPIRED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) is None

    # But cancel reject request does!
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.CREATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.NEW) == FOrdStatus.NEW
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.PARTIALLY_FILLED) == FOrdStatus.PARTIALLY_FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.FILLED) == FOrdStatus.FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.DONE_FOR_DAY) == FOrdStatus.DONE_FOR_DAY
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.CANCELED) == FOrdStatus.CANCELED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.PENDING_CANCEL) == FOrdStatus.PENDING_CANCEL
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.STOPPED) == FOrdStatus.STOPPED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.REJECTED) == FOrdStatus.REJECTED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.SUSPENDED) == FOrdStatus.SUSPENDED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.PENDING_NEW) == FOrdStatus.PENDING_NEW
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.CALCULATED) == FOrdStatus.CALCULATED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.EXPIRED) == FOrdStatus.EXPIRED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_CANCEL, '9', 0, FOrdStatus.PENDING_REPLACE) == FOrdStatus.PENDING_REPLACE
    # fmt: on


def test_state_transition__pendingreplce__execution_report():
    # fmt: off
    # Executin report doesn't not have any effect of pending cancelled state
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.CREATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.PARTIALLY_FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.FILLED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.DONE_FOR_DAY) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.CANCELED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.PENDING_CANCEL) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.STOPPED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.REJECTED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.SUSPENDED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.PENDING_NEW) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.CALCULATED) is None
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.EXPIRED) is None
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.TRADE, FOrdStatus.PENDING_REPLACE) is None
    # fmt: on


def test_state_transition__pendingreplce__execution_report_exectype_replace():
    # fmt: off
    # Executin report doesn't not have any effect of pending cancelled state
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.CREATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.NEW) == FOrdStatus.NEW
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.PARTIALLY_FILLED) == FOrdStatus.PARTIALLY_FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.FILLED) == FOrdStatus.FILLED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.DONE_FOR_DAY) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.CANCELED) == FOrdStatus.CANCELED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.PENDING_CANCEL) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.STOPPED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.REJECTED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.SUSPENDED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.PENDING_NEW) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.CALCULATED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.EXPIRED) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '8', FExecType.REPLACED, FOrdStatus.PENDING_REPLACE) 
    # fmt: on


def test_state_transition__pendingreplce__ord_reject():
    # fmt: off
    # But cancel reject request does!
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.CREATED) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.NEW) == FOrdStatus.NEW
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.PARTIALLY_FILLED) == FOrdStatus.PARTIALLY_FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.FILLED) == FOrdStatus.FILLED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.DONE_FOR_DAY) == FOrdStatus.DONE_FOR_DAY
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.CANCELED) == FOrdStatus.CANCELED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.PENDING_CANCEL) == FOrdStatus.PENDING_CANCEL
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.STOPPED) == FOrdStatus.STOPPED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.REJECTED) == FOrdStatus.REJECTED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.SUSPENDED) == FOrdStatus.SUSPENDED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.PENDING_NEW) == FOrdStatus.PENDING_NEW
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.CALCULATED) == FOrdStatus.CALCULATED
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.EXPIRED) == FOrdStatus.EXPIRED
    assert pytest.raises(FIXError, FIXNewOrderSingle.change_status, FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.ACCEPTED_FOR_BIDDING) 
    assert FIXNewOrderSingle.change_status(FOrdStatus.PENDING_REPLACE, '9', 0, FOrdStatus.PENDING_REPLACE) == FOrdStatus.PENDING_REPLACE
    # fmt: on


def test_exec_sequence__vanilla_fill():
    """
    A.1.a – Filled order
    https://www.fixtrading.org/online-specification/order-state-changes/#a-vanilla-1

    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )
    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"
    assert o.order_id is None
    assert not o.can_cancel()
    assert not o.can_replace()
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1
    assert o.order_id is not None
    assert o.status == FOrdStatus.PENDING_NEW, f"o.status={chr(o.status)}"
    assert not o.can_cancel()
    assert not o.can_replace()
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.NEW, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 10

    assert o.can_cancel() > 0
    assert o.can_replace() > 0
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=2,
        leaves_qty=8,
        last_qty=2,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PARTIALLY_FILLED, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 2
    assert o.leaves_qty == 8

    assert o.can_cancel() > 0
    assert o.can_replace() > 0
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=3,
        leaves_qty=7,
        last_qty=1,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PARTIALLY_FILLED, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 3
    assert o.leaves_qty == 7

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.FILLED,
        cum_qty=10,
        leaves_qty=0,
        last_qty=7,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.FILLED, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 10
    assert o.leaves_qty == 0

    assert not o.can_cancel()
    assert not o.can_replace()
    assert o.is_finished() == 1


def test_exec_sequence__vanilla_fill_reject__pendingnew():
    """
    A.1.a – Filled ordern (reject)
    https://www.fixtrading.org/online-specification/order-state-changes/#a-vanilla-1

    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PENDING_NEW, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.REJECTED, FOrdStatus.REJECTED, cum_qty=0, leaves_qty=0
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.REJECTED, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 00

    assert not o.can_cancel()
    assert not o.can_replace()
    assert o.is_finished() == 1


def test_exec_sequence__vanilla_fill__reject_new():
    """
    A.1.a – Filled order (reject when new confirmed)
    https://www.fixtrading.org/online-specification/order-state-changes/#a-vanilla-1

    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PENDING_NEW, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.NEW, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 10

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.REJECTED, FOrdStatus.REJECTED, cum_qty=0, leaves_qty=0
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.REJECTED, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 00


def test_exec_sequence__vanilla_suspended():
    """
    A.1.b – Part-filled day order, done for day -> suspended
    https://www.fixtrading.org/online-specification/order-state-changes/#a-vanilla-1

    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PENDING_NEW, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.NEW, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 10

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=2,
        leaves_qty=8,
        last_qty=2,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PARTIALLY_FILLED, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 2
    assert o.leaves_qty == 8

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.SUSPENDED,
        FOrdStatus.SUSPENDED,
        cum_qty=2,
        leaves_qty=0,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.SUSPENDED, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 2
    assert o.leaves_qty == 0

    assert o.can_cancel()
    assert o.can_replace()
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.NEW,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=2,
        leaves_qty=8,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PARTIALLY_FILLED, f"o.status={chr(o.status)}"
    assert o.qty == 10
    assert o.cum_qty == 2
    assert o.leaves_qty == 8


def test_cancel_req():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.SELL, price=100.0, qty=20
    )
    o.status = FOrdStatus.NEW

    m = o.cancel_req()
    FIX_SCHEMA.validate(m)

    # Tag 11: ClOrdID
    assert m[11] == "clordTest-1"

    # Tag 41: OrigClOrdID
    assert m[41] == o.clord_id

    # Tag 38: Order Qty
    assert m[38] == "20"

    # Tag 54: Side
    assert m[54] == "2"  # sell

    #  Tag 55: Symbol set to v2 symbol
    assert m[FTag.Symbol] == "US.F.TICKER"

    # Overall message is valid!
    assert FIX_SCHEMA.validate(m)


def test_cancel_req__zero_filled_order():
    """
    B.1.a – Cancel request issued for a zero-filled order
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_cxl_request(o)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.CANCELED, FOrdStatus.CANCELED, cum_qty=0, leaves_qty=0
    )

    with pytest.raises(FIXError, match="incorrect message type"):
        o.process_execution_report(cxl_req)

    assert o.process_execution_report(msg) == 1
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 0

    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished() == 1


def test_cancel_req__zero_filled_order__cancel_reject():
    """
    B.1.a – Cancel request issued for a zero-filled order
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.NEW

    cxl_req = ft.fix_cxl_request(o)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    msg = ft.fix_cxlrep_reject_msg(cxl_req, FOrdStatus.NEW)
    assert msg.msg_type == "9"
    rc = o.process_cancel_rej_report(msg)
    assert rc == 1, rc
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 10
    assert o.status == FOrdStatus.NEW

    assert o.can_replace()
    assert o.can_cancel()
    assert o.is_finished() == 0


def test_cancel_req__zero_filled_order__cancel_reject_after_pending():
    """
    B.1.a – Cancel request issued for a zero-filled order
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_cxl_request(o)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.PENDING_CANCEL,
        FOrdStatus.PENDING_CANCEL,
        cum_qty=0,
        leaves_qty=10,
    )
    assert o.process_execution_report(msg) == 0  # Just ignored!
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 10

    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    msg = ft.fix_cxlrep_reject_msg(cxl_req, FOrdStatus.NEW)
    assert msg.msg_type == "9"
    rc = o.process_cancel_rej_report(msg)
    assert rc == 1, rc
    assert o.qty == 10
    assert o.cum_qty == 0
    assert o.leaves_qty == 10
    assert o.status == FOrdStatus.NEW

    assert o.can_replace()
    assert o.can_cancel()
    assert o.is_finished() == 0


def test_cancel_req__part_filled_order__with_some_execution_between():
    """
    B.1.b – Cancel request issued for a part-filled order – executions occur whilst cancel request is active
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=2,
        leaves_qty=8,
        last_qty=2,
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_cxl_request(o)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=5,
        leaves_qty=5,
        last_qty=3,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    assert o.qty == 10
    assert o.cum_qty == 5
    assert o.leaves_qty == 5

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.PENDING_CANCEL,
        FOrdStatus.PENDING_CANCEL,
        cum_qty=5,
        leaves_qty=5,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    assert o.qty == 10
    assert o.cum_qty == 5
    assert o.leaves_qty == 5

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PENDING_CANCEL,
        cum_qty=6,
        leaves_qty=4,
        last_qty=1,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    assert o.qty == 10
    assert o.cum_qty == 6
    assert o.leaves_qty == 4

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.CANCELED,
        FOrdStatus.CANCELED,
        cum_qty=6,
        leaves_qty=0,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.CANCELED
    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished() == 1
    assert o.qty == 10
    assert o.cum_qty == 6
    assert o.leaves_qty == 0


def test_cancel_req__order_filled_before_cancel_accepted_different_clord():
    """
    B.1.c – Cancel request issued for an order that becomes filled before cancel request can be accepted
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=2,
        leaves_qty=8,
        last_qty=2,
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_cxl_request(o)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    assert o.orig_clord_id
    assert o.clord_id != o.orig_clord_id
    # IMPORTANT: USING OLD CLORD because pretending this report was generated
    # before request for cancel arrived to server
    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=5,
        leaves_qty=5,
        last_qty=3,
    )
    rc = o.process_execution_report(msg)
    assert rc == 0, rc
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    assert o.qty == 10
    assert o.cum_qty == 5
    assert o.leaves_qty == 5

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.PENDING_CANCEL,
        FOrdStatus.PENDING_CANCEL,
        cum_qty=5,
        leaves_qty=5,
        last_qty=nan,
        price=nan,
        order_qty=nan,
        orig_clord_id=o.orig_clord_id,
    )
    assert o.process_execution_report(msg) == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,
        FExecType.TRADE,
        FOrdStatus.PENDING_CANCEL,
        cum_qty=10,
        leaves_qty=0,
        last_qty=5,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    assert o.qty == 10
    assert o.cum_qty == 10
    assert o.leaves_qty == 0

    msg = ft.fix_cxlrep_reject_msg(cxl_req, FOrdStatus.FILLED)
    assert o.process_cancel_rej_report(msg) == 1
    assert o.qty == 10
    assert o.cum_qty == 10
    assert o.leaves_qty == 0
    assert o.status == FOrdStatus.FILLED
    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished() == 1


def test_cancel_req__not_acknoledged_order_by_gate():
    """
    B.1.c – Cancel request issued for an order that becomes filled before cancel request can be accepted
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    with pytest.raises(FIXError, match="order is not allowed for cancel"):
        cxl_req = o.cancel_req()
    assert not o.can_replace()
    assert not o.can_cancel()

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    with pytest.raises(FIXError, match="order is not allowed for cancel"):
        cxl_req = o.cancel_req()
    assert not o.can_replace()
    assert not o.can_cancel()


def test_cancel_req__multiple_requests_are_blocked():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_cxl_request(o)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0


def test_cancel_req__order_filled_before_cancel_accepted():
    """
    B.1.c – Cancel request issued for an order that becomes filled before cancel request can be accepted
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=2,
        leaves_qty=8,
        last_qty=2,
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_cxl_request(o)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    msg = ft.fix_cxlrep_reject_msg(cxl_req, FOrdStatus.REJECTED)
    assert o.process_cancel_rej_report(msg) == 1
    assert o.qty == 10
    assert o.cum_qty == 2
    assert o.leaves_qty == 0
    assert o.status == FOrdStatus.REJECTED
    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished()


def test_replace_req():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.SELL, price=100.0, qty=20
    )
    o.status = FOrdStatus.NEW

    old_clord = o.clord_id
    m = o.replace_req(200, 30)
    FIX_SCHEMA.validate(m)

    # Tag 11: ClOrdID
    assert m[11] == "clordTest-1"

    # Tag 41: OrigClOrdID
    assert m[41] == old_clord

    # Tag 38: Order Qty
    assert m[38] == "30"

    # Tag 44: Order Price
    assert m[44] == "200"

    # Tag 54: Side
    assert m[54] == "2"  # sell

    #  Tag 55: Symbol set to v2 symbol
    assert m[55] == "US.F.TICKER"

    # Overall message is valid!
    assert FIX_SCHEMA.validate(m)


def test_replace_req_only_price():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.SELL, price=100.0, qty=20
    )
    o.status = FOrdStatus.NEW

    m = o.replace_req(200, o.qty)
    FIX_SCHEMA.validate(m)

    # Tag 38: Order Qty
    assert m[38] == "20"

    # Tag 44: Order Price
    assert m[44] == "200"


def test_replace_req_only_qty():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.SELL, price=100.0, qty=20
    )
    o.status = FOrdStatus.NEW

    m = o.replace_req(nan, 30)
    FIX_SCHEMA.validate(m)

    # Tag 38: Order Qty
    assert m[38] == "30"

    # Tag 44: Order Price
    assert m[44] == "100.0"


def test_replace_req__not_set():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.SELL, price=100.0, qty=20
    )
    o.status = FOrdStatus.NEW

    with pytest.raises(FIXError, match="no price / qty change in replace_req"):
        m = o.replace_req(nan, nan)

    # No change in price/qty
    with pytest.raises(FIXError, match="no price / qty change in replace_req"):
        m = o.replace_req(o.price, o.qty)

    # No change in price/qty
    with pytest.raises(FIXError, match="no price / qty change in replace_req"):
        m = o.replace_req(o.price, 0)


def test_replace_req__zero_filled__increased_qty():
    """
    C.1.a – Zero-filled order, cancel/replace request issued to increase order qty
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_rep_request(o, 300, 11)
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10

    with pytest.raises(FIXError, match="Order cannot be replaced"):
        cxl_req = ft.fix_rep_request(o, 300, 11)

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.PENDING_REPLACE,
        FOrdStatus.PENDING_REPLACE,
        cum_qty=0,
        leaves_qty=10,
    )
    assert o.process_execution_report(msg) == 0
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.REPLACED,
        FOrdStatus.NEW,
        cum_qty=0,
        leaves_qty=11,
        last_qty=nan,
        price=300,
        order_qty=11,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.NEW
    assert o.can_replace() > 0
    assert o.can_cancel() > 0
    assert o.is_finished() == 0
    assert o.price == 300
    assert o.qty == 11
    assert o.orig_clord_id is None


def test_replace_req__part_filled__increased_qty_while_pending_replace_fractional_fill():
    """
    C.1.b – Part-filled order, followed by cancel/replace request to increase order qty, execution occurs whilst order is pending replace
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=1,
        leaves_qty=9,
        last_qty=1,
    )
    assert o.process_execution_report(msg) == 1
    old_clord = o.clord_id

    cxl_req = ft.fix_rep_request(o, 300, 12)
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.PENDING_REPLACE,
        FOrdStatus.PENDING_REPLACE,
        cum_qty=1,
        leaves_qty=9,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10

    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,  # ORIG!!!
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=1.1,
        leaves_qty=8.9,
        last_qty=0.1,
    )
    assert o.process_execution_report(msg) == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.REPLACED,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=1.1,
        leaves_qty=10.9,
        last_qty=nan,
        price=300,
        order_qty=12,
        orig_clord_id=o.orig_clord_id,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PARTIALLY_FILLED
    assert o.can_replace() > 0
    assert o.can_cancel() > 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 300
    assert o.qty == 12
    assert o.cum_qty == 1.1
    assert o.leaves_qty == 10.9
    assert o.orig_clord_id is None
    assert o.clord_id != old_clord

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.FILLED,
        cum_qty=12,
        leaves_qty=0,
        last_qty=10.9,
    )
    assert o.process_execution_report(msg) == 1

    with pytest.raises(FIXError, match="incorrect message type"):
        o.process_cancel_rej_report(msg)

    assert o.status == FOrdStatus.FILLED
    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished() == 1
    # Don't change order price/qty until confirmed
    assert o.price == 300
    assert o.qty == 12
    assert o.cum_qty == 12
    assert o.leaves_qty == 0


def test_replace_req__zero_filled__cxlrep_reject_when_new():
    """
    C.1.a – Zero-filled order, cancel/replace request issued, but rejected back to new state
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_rep_request(o, 300, 11)
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10

    msg = ft.fix_cxlrep_reject_msg(cxl_req, FOrdStatus.NEW)
    assert o.process_cancel_rej_report(msg) == 1
    assert o.can_replace() > 0
    assert o.can_cancel() > 0
    assert o.is_finished() == 0

    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10


def test_replace_req__filled_order_rejected_after_filled():
    """
    C.1.c – Filled order, followed by cancel/replace request to increase order quantity
    (CASE 1: reject after fill)
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_rep_request(o, 300, 12)
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10

    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,
        FExecType.TRADE,
        FOrdStatus.FILLED,
        cum_qty=10,
        leaves_qty=0,
        last_qty=10,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10
    assert o.cum_qty == 10
    assert o.leaves_qty == 0

    msg = ft.fix_cxlrep_reject_msg(cxl_req, FOrdStatus.FILLED)
    assert o.process_cancel_rej_report(msg) == 1
    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished() == 1

    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10
    assert o.cum_qty == 10
    assert o.leaves_qty == 0


def test_replace_req__filled_order_rejected__filled_increase_passed():
    """
    C.1.c – Filled order, followed by cancel/replace request to increase order quantity
    (CASE 2: fill passed, but following qty increase also passed)
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_rep_request(o, 300, 12)
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10

    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,
        FExecType.TRADE,
        FOrdStatus.FILLED,
        cum_qty=10,
        leaves_qty=0,
        last_qty=10,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10
    assert o.cum_qty == 10
    assert o.leaves_qty == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.PENDING_REPLACE,
        FOrdStatus.PENDING_REPLACE,
        cum_qty=10,
        leaves_qty=0,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.REPLACED,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=10,
        leaves_qty=2,
        last_qty=nan,
        price=300,
        order_qty=12,
        orig_clord_id=o.orig_clord_id,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PARTIALLY_FILLED
    assert o.can_replace() > 0
    assert o.can_cancel() > 0
    assert o.is_finished() == 0

    assert o.price == 300
    assert o.qty == 12
    assert o.cum_qty == 10
    assert o.leaves_qty == 2

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.FILLED,
        cum_qty=12,
        leaves_qty=0,
        last_qty=2,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.FILLED

    assert o.price == 300
    assert o.qty == 12
    assert o.cum_qty == 12
    assert o.leaves_qty == 0


def test_replace_req__replace_price_only_but_rejected_because_fill():
    """
    C.2.a – Cancel/replace request (not for quantity change) is rejected as a fill has occurred
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_rep_request(o, 300, nan)
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,
        FExecType.TRADE,
        FOrdStatus.FILLED,
        cum_qty=10,
        leaves_qty=0,
        last_qty=10,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_REPLACE
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0
    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10
    assert o.cum_qty == 10
    assert o.leaves_qty == 0

    msg = ft.fix_cxlrep_reject_msg(cxl_req, FOrdStatus.FILLED)
    assert o.process_cancel_rej_report(msg) == 1
    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished() == 1

    # Don't change order price/qty until confirmed
    assert o.price == 200
    assert o.qty == 10
    assert o.cum_qty == 10
    assert o.leaves_qty == 0


def test_replace_req__decreased_qty():
    """
    C.3.a – Cancel/replace request sent whilst execution is being reported –
    the requested order qty exceeds the cum qty. Order is replaced then filled
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=2,
        leaves_qty=8,
        last_qty=2,
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_rep_request(o, nan, 9)
    assert o.status == FOrdStatus.PENDING_REPLACE

    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,
        FExecType.TRADE,
        FOrdStatus.PENDING_REPLACE,
        cum_qty=3,
        leaves_qty=7,
        last_qty=1,
    )
    assert o.process_execution_report(msg) == 0
    assert o.status == FOrdStatus.PENDING_REPLACE

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.REPLACED,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=3,
        leaves_qty=6,
        last_qty=nan,
        price=200,
        order_qty=9,
        orig_clord_id=o.orig_clord_id,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.PARTIALLY_FILLED
    assert o.can_replace() > 0
    assert o.can_cancel() > 0
    assert o.is_finished() == 0

    assert o.price == 200
    assert o.qty == 9
    assert o.cum_qty == 3
    assert o.leaves_qty == 6


def test_replace_req__decreased_qty_exact_match_to_fill():
    """
    C.3.b – Cancel/replace request sent whilst execution is being reported –
    the requested order qty equals the cum qty – order qty is amended to cum qty
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_rep_request(o, nan, 7)
    assert o.status == FOrdStatus.PENDING_REPLACE

    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=7,
        leaves_qty=3,
        last_qty=7,
    )
    assert o.process_execution_report(msg) == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.REPLACED,
        FOrdStatus.FILLED,
        cum_qty=7,
        leaves_qty=0,
        last_qty=nan,
        price=200,
        order_qty=7,
        orig_clord_id=o.orig_clord_id,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.FILLED
    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished() == 1

    assert o.price == 200
    assert o.qty == 7
    assert o.cum_qty == 7
    assert o.leaves_qty == 0


def test_replace_req__decreased_qty__also_less_than_cum_qty():
    """
    C.3.c – Cancel/replace request sent whilst execution is being reported –
    the requested order qty is below cum qty – order qty is amended to cum qty
    :return:
    """
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1

    cxl_req = ft.fix_rep_request(o, nan, 7)
    assert o.status == FOrdStatus.PENDING_REPLACE

    msg = ft.fix_exec_report_msg(
        o,
        o.orig_clord_id,
        FExecType.TRADE,
        FOrdStatus.PARTIALLY_FILLED,
        cum_qty=8,
        leaves_qty=2,
        last_qty=8,
    )
    assert o.process_execution_report(msg) == 0

    msg = ft.fix_exec_report_msg(
        o,
        o.clord_id,
        FExecType.REPLACED,
        FOrdStatus.FILLED,
        cum_qty=8,
        leaves_qty=0,
        last_qty=nan,
        price=200,
        order_qty=8,
        orig_clord_id=o.orig_clord_id,
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.FILLED
    assert not o.can_replace()
    assert not o.can_cancel()
    assert o.is_finished() == 1

    assert o.price == 200
    assert o.qty == 8
    assert o.cum_qty == 8
    assert o.leaves_qty == 0


def test_exec_report_clord_mismatch():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, "unknown clord", FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    with pytest.raises(FIXError, match="orig_clord_id mismatch"):
        o.process_execution_report(msg)

    o.orig_clord_id = "1234"
    msg = ft.fix_exec_report_msg(
        o, "1234", FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    o.process_execution_report(msg)


def test_cancel_req__cancel_reject_invalid_transition():
    o = FIXNewOrderSingle(
        "clordTest", "US.F.TICKER", side=FOrdSide.BUY, price=200.0, qty=10
    )

    ft = FIXTester(FIX_SCHEMA)
    assert ft.order_register_single(o) == 1
    assert o.status == FOrdStatus.CREATED, f"o.status={chr(o.status)}"

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.PENDING_NEW, FOrdStatus.PENDING_NEW
    )
    assert o.process_execution_report(msg) == 1

    msg = ft.fix_exec_report_msg(
        o, o.clord_id, FExecType.NEW, FOrdStatus.NEW, cum_qty=0, leaves_qty=10
    )
    assert o.process_execution_report(msg) == 1
    assert o.status == FOrdStatus.NEW

    cxl_req = ft.fix_cxl_request(o)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0

    msg = ft.fix_cxlrep_reject_msg(cxl_req, FOrdStatus.ACCEPTED_FOR_BIDDING)

    assert not o.process_cancel_rej_report(msg)
    assert o.status == FOrdStatus.PENDING_CANCEL
    assert o.can_replace() == 0
    assert o.can_cancel() == 0
    assert o.is_finished() == 0



import pickle
import sqlite3

from asyncfix.message import FIXMessage, MessageDirection
from asyncfix.session import FIXSession


class DuplicateSeqNoError(Exception):
    pass


class Journaler(object):
    def __init__(self, filename=None):
        if filename is None:
            self.conn = sqlite3.connect(":memory:")
        else:
            self.conn = sqlite3.connect(filename)

        self.cursor = self.conn.cursor()
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS message("
            "seqNo INTEGER NOT NULL,"
            "session TEXT NOT NULL,"
            "direction INTEGER NOT NULL,"
            "msg TEXT,"
            "PRIMARY KEY (seqNo, session, direction))"
        )

        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS session("
            "sessionId INTEGER PRIMARY KEY AUTOINCREMENT,"
            "targetCompId TEXT NOT NULL,"
            "senderCompId TEXT NOT NULL,"
            "outboundSeqNo INTEGER DEFAULT 0,"
            "inboundSeqNo INTEGER DEFAULT 0,"
            "UNIQUE (targetCompId, senderCompId))"
        )

    def sessions(self):
        sessions = []
        self.cursor.execute(
            "SELECT sessionId, targetCompId, senderCompId, outboundSeqNo, inboundSeqNo"
            " FROM session"
        )
        for sessionInfo in self.cursor:
            session = FIXSession(sessionInfo[0], sessionInfo[1], sessionInfo[2])
            session.snd_seq_num = sessionInfo[3]
            session.next_expected_msg_seq_num = sessionInfo[4] + 1
            sessions.append(session)

        return sessions

    def create_session(self, target_comp_id, sender_comp_id) -> FIXSession:
        session = None
        try:
            self.cursor.execute(
                "INSERT INTO session(targetCompId, senderCompId) VALUES(?, ?)",
                (target_comp_id, sender_comp_id),
            )
            sessionId = self.cursor.lastrowid
            self.conn.commit()
            session = FIXSession(sessionId, target_comp_id, sender_comp_id)
        except sqlite3.IntegrityError:
            raise RuntimeError(
                "Session already exists for TargetCompId: %s SenderCompId: %s"
                % (target_comp_id, sender_comp_id)
            )

        return session

    def persist_msg(
        self, msg: FIXMessage, session: FIXSession, direction: MessageDirection
    ):
        msgStr = pickle.dumps(msg)
        seqNo = msg["34"]
        try:
            self.cursor.execute(
                "INSERT INTO message VALUES(?, ?, ?, ?)",
                (seqNo, session.key, direction.value, msgStr),
            )
            if direction == MessageDirection.OUTBOUND:
                self.cursor.execute("UPDATE session SET outboundSeqNo=?", (seqNo,))
            elif direction == MessageDirection.INBOUND:
                self.cursor.execute("UPDATE session SET inboundSeqNo=?", (seqNo,))

            self.conn.commit()
        except sqlite3.IntegrityError as e:
            raise DuplicateSeqNoError("%s is a duplicate, error %s" % (seqNo, repr(e)))

    def recover_msg(self, session: FIXSession, direction: MessageDirection, seq_no):
        try:
            msgs = self.recover_msgs(session, direction, seq_no, seq_no)
            return msgs[0]
        except IndexError:
            return None

    def recover_msgs(
        self, session: FIXSession, direction: MessageDirection, start_seq_no, end_seq_no
    ):
        self.cursor.execute(
            "SELECT msg FROM message WHERE session = ? AND direction = ? AND seqNo >= ?"
            " AND seqNo <= ? ORDER BY seqNo",
            (session.key, direction.value, start_seq_no, end_seq_no),
        )
        msgs = []
        for msg in self.cursor:
            msgs.append(pickle.loads(msg[0]))
        return msgs

    def get_all_msgs(
        self,
        sessions: list[FIXSession] | None = None,
        direction: MessageDirection | None = None,
    ):
        sql = "SELECT seqNo, msg, direction, session FROM message"
        clauses = []
        args = []
        if sessions is not None and len(sessions) != 0:
            clauses.append("session in (" + ",".join("?" * len(sessions)) + ")")
            args.extend(sessions)
        if direction is not None:
            clauses.append("direction = ?")
            args.append(direction.value)

        if clauses:
            sql = sql + " WHERE " + " AND ".join(clauses)

        sql = sql + " ORDER BY rowid"

        self.cursor.execute(sql, tuple(args))
        msgs = []
        for msg in self.cursor:
            msgs.append((msg[0], pickle.loads(msg[1]), msg[2], msg[3]))

        return msgs



import asyncio
import logging

from asyncfix.connection import ConnectionState, FIXConnectionHandler, FIXEndPoint
from asyncfix.engine import FIXEngine
from asyncfix.journaler import DuplicateSeqNoError
from asyncfix.message import FIXMessage
from asyncfix.protocol import FIXProtocolBase


class FIXServerConnectionHandler(FIXConnectionHandler):
    def __init__(
        self,
        engine: FIXEngine,
        protocol: FIXProtocolBase,
        socket_reader: asyncio.StreamReader,
        socket_writer: asyncio.StreamWriter,
        addr=None,
        observer=None,
    ):
        FIXConnectionHandler.__init__(
            self, engine, protocol, socket_reader, socket_writer, addr, observer
        )

    async def handle_session_message(self, msg: FIXMessage):
        protocol = self.codec.protocol

        recv_seq_no = msg[protocol.fixtags.MsgSeqNum]

        msg_type = msg[protocol.fixtags.MsgType]
        target_comp_id = msg[protocol.fixtags.TargetCompID]
        sender_comp_id = msg[protocol.fixtags.SenderCompID]
        responses = []

        if msg_type == protocol.msgtype.LOGON:
            if self.connection_state == ConnectionState.LOGGED_IN:
                logging.warning(
                    "Client session already logged in - ignoring login request"
                )
            else:
                # compids are reversed here...
                self.session = self.engine.get_or_create_session_from_comp_ids(
                    sender_comp_id, target_comp_id
                )
                if self.session is not None:
                    try:
                        self.connection_state = ConnectionState.LOGGED_IN
                        self.heartbeat_period = float(msg[protocol.fixtags.HeartBtInt])
                        responses.append(protocol.logon())
                    except DuplicateSeqNoError:
                        logging.error(
                            "Failed to process login request with duplicate seq no"
                        )
                        await self.disconnect()
                        return
                else:
                    logging.warning(
                        "Rejected login attempt for invalid session (SenderCompId: %s,"
                        " TargetCompId: %s)" % (sender_comp_id, target_comp_id)
                    )
                    await self.disconnect()
                    return  # we have to return here since self.session won't be valid
        elif self.connection_state == ConnectionState.LOGGED_IN:
            # compids are reversed here
            if not self.session.validate_comp_ids(sender_comp_id, target_comp_id):
                logging.error("Received message with unexpected comp ids")
                await self.disconnect()
                return

            if msg_type == protocol.msgtype.LOGOUT:
                self.connection_state = ConnectionState.LOGGED_OUT
                self.handle_close()
            elif msg_type == protocol.msgtype.TESTREQUEST:
                responses.append(protocol.heartbeat())
            elif msg_type == protocol.msgtype.RESENDREQUEST:
                responses.extend(self._handle_resend_request(msg))
            elif msg_type == protocol.msgtype.SEQUENCERESET:
                new_seq_no = msg[protocol.fixtags.NewSeqNo]
                self.session.set_recv_seq_no(int(new_seq_no) - 1)
                recv_seq_no = new_seq_no
        else:
            logging.warning("Can't process message, counterparty is not logged in")

        return (recv_seq_no, responses)


class FIXServer(FIXEndPoint):
    def __init__(self, engine: FIXEngine, protocol: FIXProtocolBase):
        FIXEndPoint.__init__(self, engine, protocol)
        self.server = None

    async def start(self, host: str, port: int):
        self.connections = []
        self.server = await asyncio.start_server(self.handle_accept, host, port)

        logging.debug("Awaiting Connections " + host + ":" + str(port))
        async with self.server:
            await self.server.serve_forever()

    async def handle_accept(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        try:
            addr = writer.get_extra_info("peername")
            logging.info("Connection from %s" % repr(addr))

            connection = FIXServerConnectionHandler(
                self.engine, self.protocol, reader, writer, addr, self
            )

            asyncio.create_task(connection.handle_read())

            self.connections.append(connection)
            for handler in filter(
                lambda x: x[1] == ConnectionState.CONNECTED, self.message_handlers
            ):
                await handler[0](connection)
        except Exception:
            logging.exception("handle_accept")



from asyncfix import FIXMessage, FMsg, FTag


class FIXProtocolBase(object):
    beginstring: str = "FIXProtocolBase"
    repeating_groups: dict[str, list[str]] = {}
    session_message_types: set = set()
    fixtags: FTag = FTag
    msgtype: FMsg = FMsg

    def logon(self) -> FIXMessage:
        raise NotImplementedError("Implement this method in child class")

    def logout(self) -> FIXMessage:
        raise NotImplementedError("Implement this method in child class")

    def heartbeat(self) -> FIXMessage:
        raise NotImplementedError("Implement this method in child class")

    def test_request(self) -> FIXMessage:
        raise NotImplementedError("Implement this method in child class")

    def sequence_reset(self, responding_to, is_gap_fill) -> FIXMessage:
        raise NotImplementedError("Implement this method in child class")

    def resend_request(self, begin_seq_no, end_seq_no="0") -> FIXMessage:
        raise NotImplementedError("Implement this method in child class")



import os
import xml.etree.ElementTree as ET
from unittest.mock import patch

import pytest

from asyncfix import FIXMessage, FMsg, FTag
from asyncfix.errors import FIXMessageError
from asyncfix.protocol.schema import (
    FIXSchema,
    SchemaComponent,
    SchemaField,
    SchemaGroup,
    SchemaMessage,
    SchemaSet,
)

TEST_DIR = os.path.abspath(os.path.dirname(__file__))


@pytest.fixture
def fix_simple_xml():
    return ET.parse(os.path.join(TEST_DIR, "schema_fix_simple.xml"))


@pytest.fixture
def fix_circular_invalid_xml():
    return ET.parse(os.path.join(TEST_DIR, "schema_fix_comp_circular.xml"))


def test_schema_field():
    f = SchemaField("1", "Account", "STRING")
    f2 = SchemaField("11", "ClOrdID", "STRING")
    assert f.tag == "1"
    assert f.name == "Account"
    assert f.ftype == "STRING"
    assert f.values == {}
    # no values in SchemaField any allowed
    assert f.validate_value("adslkajdlskj")

    f.values["test"] = "1"
    assert f.values == {"test": "1"}
    with pytest.raises(
        FIXMessageError, match="Field=Account value expected to be one of the "
    ):
        assert not f.validate_value("adslkajdlskj")
    assert f.validate_value("test")

    assert f != f2
    assert f == f
    assert f == "Account"
    assert f == "1"
    assert f == 1
    assert f != float("123")

    fields = {f: 1, f2: 2}

    assert f in fields
    assert fields[f] == 1
    assert fields[f2] == 2
    assert fields["Account"] == 1

    # search by tag not supported
    assert "1" not in fields
    assert 1 not in fields


def test_schema_set():
    f = SchemaField("1", "Account", "STRING")
    f2 = SchemaField("11", "ClOrdID", "STRING")

    s = SchemaSet("myset")
    s.add(f, True)
    s.add(f2, False)
    assert hash(s) == hash("myset")

    with pytest.raises(ValueError, match="Unsupported field_or_set type, got"):
        s.add("notsupported", False)

    with pytest.raises(ValueError, match="tag property is not supported"):
        s.tag

    assert f in s
    assert f2 in s
    assert "Account" in s

    assert s.name == "myset"
    assert len(s.members) == 2
    assert s.required[f]
    assert not s.required[f2]

    # Error type for group
    fg = SchemaField("73", "NoOrders", "STRING")
    with pytest.raises(
        ValueError, match="SchemaSet expected to have group field with "
    ):
        SchemaSet("NoOrders", field=fg)

    with pytest.raises(FIXMessageError, match="item looks like tag"):
        assert "1" in s

    with pytest.raises(FIXMessageError, match="item looks like tag"):
        assert 1 in s

    with pytest.raises(FIXMessageError, match="item looks like tag"):
        s["1"]

    with pytest.raises(FIXMessageError, match="item looks like tag"):
        s[1]


def test_schema_set_included_components():
    f = SchemaField("1", "Account", "STRING")
    f2 = SchemaField("11", "ClOrdID", "STRING")

    s = SchemaSet("myset")
    s.add(f, True)

    s2 = SchemaSet("myset2")
    s2.add(f2, False)

    s.merge(s2)

    assert f in s
    assert f2 in s
    assert s == "myset"

    assert s.name == "myset"
    assert len(s.members) == 2
    assert s.required[f]
    assert not s.required[f2]


def test_schema_set__add_group():
    f = SchemaField("1", "Account", "STRING")
    f2 = SchemaField("11", "ClOrdID", "STRING")

    s = SchemaSet("myset")
    s.add(f, True)

    fg = SchemaField("73", "NoOrders", "NUMINGROUP")
    g = SchemaGroup(fg, False)
    assert g.tag == fg.tag
    assert not g.field_required
    assert g.field is fg
    assert g == fg
    assert g == "NoOrders"
    assert g == "73"
    assert g == 73

    s2 = SchemaSet("myset2")
    s2.add(g, False)

    s.merge(s2)

    assert s.name == "myset"
    assert len(s.members) == 2
    assert s.members[f]
    assert isinstance(s.members[f], SchemaField)
    assert not s.required[g]

    assert f in s
    assert g in s
    assert "Account" in s
    assert "NoOrders" in s, s.members
    assert fg in s


def test_xml_init(fix_simple_xml):
    schema = FIXSchema(fix_simple_xml)

    f = schema.field2tag["AdvSide"]
    f2 = schema.tag2field["4"]

    assert f == f2
    assert f.validate_value("B")
    with pytest.raises(
        FIXMessageError, match="Field=AdvSide value expected to be one of the "
    ):
        assert not f.validate_value("Z")

    assert len(schema.components) == 3
    assert "ContraGrp" in schema.components
    assert "ContraGrp2" in schema.components
    assert "CommissionData" in schema.components
    c = schema.components["ContraGrp"]
    assert isinstance(c, SchemaComponent)
    assert c.name == "ContraGrp"
    assert c.field is None

    # Check messages
    assert schema.messages
    m = schema.messages["ExecutionReport"]
    assert isinstance(m, SchemaMessage)
    assert m.msg_type == "8"
    assert m.msg_cat == "app"
    assert isinstance(m["NoPartyIDs"], SchemaGroup)

    # Ensure group ordered!
    g = m["NoPartyIDs"]
    assert isinstance(g, SchemaGroup)
    assert ["PartyID", "PartyRole"] == list(g.members)


def test_xml_circular_init(fix_circular_invalid_xml):
    with pytest.raises(
        RuntimeError,
        match=(
            r"Failed to resolve component circular reference.Failed components:"
            r" \['ContraGrp'\]"
        ),
    ):
        schema = FIXSchema(fix_circular_invalid_xml)


def test_xml_real_schema():
    tree = ET.parse(os.path.join(TEST_DIR, "FIX44.xml"))
    schema = FIXSchema(tree)

    assert len(schema.messages) == 93
    assert len(schema.components) == 104
    assert len(schema.field2tag) == 912
    assert schema.types == {
        "AMT",
        "BOOLEAN",
        "CHAR",
        "COUNTRY",
        "CURRENCY",
        "DATA",
        "EXCHANGE",
        "FLOAT",
        "INT",
        "LENGTH",
        "LOCALMKTDATE",
        "MONTHYEAR",
        "MULTIPLEVALUESTRING",
        "NUMINGROUP",
        "PERCENTAGE",
        "PRICE",
        "PRICEOFFSET",
        "QTY",
        "SEQNUM",
        "STRING",
        "UTCDATEONLY",
        "UTCTIMEONLY",
        "UTCTIMESTAMP",
    }

    assert "ExecutionReport" in schema.messages


def test_xml_real_schema_tt():
    tree = ET.parse(os.path.join(TEST_DIR, "TT-FIX44.xml"))
    schema = FIXSchema(tree)

    assert len(schema.messages) == 40
    assert len(schema.components) == 0
    assert len(schema.field2tag) == 611
    assert schema.types == {
        "AMT",
        "BOOLEAN",
        "CHAR",
        "CURRENCY",
        "DATA",
        "DAYOFMONTH",
        "EXCHANGE",
        "FLOAT",
        "INT",
        "LENGTH",
        "LOCALMKTDATE",
        "MONTHYEAR",
        "MULTIPLESTRINGVALUE",
        "NUMINGROUP",
        "PRICE",
        "QTY",
        "SEQNUM",
        "STRING",
        "UTCDATEONLY",
        "UTCTIMEONLY",
        "UTCTIMESTAMP",
    }

    assert "ExecutionReport" in schema.messages

    # Each message can have different group members of the same group
    m = schema.messages["MarketDataSnapshot"]
    g = m["NoMDEntries"]
    assert isinstance(g, SchemaGroup)
    assert g.keys() == [
        "MDEntryType",
        "MDEntryPx",
        "MDEntrySize",
        "MDEntryDate",
        "MDEntryTime",
        "MDEntryPositionNo",
        "NumberOfOrders",
    ]

    # Each message can have different group members of the same group
    m = schema.messages["MarketDataIncrementalRefresh"]
    g = m["NoMDEntries"]
    assert isinstance(g, SchemaGroup)
    assert g.keys() == [
        "MDUpdateAction",
        "MDEntryType",
        "Symbol",
        "SecurityDesc",
        "Product",
        "SecurityType",
        "SecuritySubType",
        "MaturityMonthYear",
        "MaturityDate",
        "MaturityDay",
        "PutOrCall",
        "StrikePrice",
        "OptAttribute",
        "DeliveryTerm",
        "DeliveryDate",
        "SecurityID",
        "SecurityExchange",
        "ExDestination",
        "CFICode",
        "Currency",
        "MDEntryPx",
        "MDEntrySize",
        "MDEntryDate",
        "MDEntryTime",
        "MDEntryPositionNo",
        "SecondaryOrderID",
        "NumberOfOrders",
    ]


def test_schema_validation(fix_simple_xml):
    schema = FIXSchema(fix_simple_xml)

    m = FIXMessage(FMsg.EXECUTIONREPORT, {FTag.OrderID: "1234"})
    schema.validate(m)

    m = FIXMessage("1ASD")
    with pytest.raises(FIXMessageError, match="msg_type=`1ASD` not in schema"):
        schema.validate(m)

    m = FIXMessage(FMsg.EXECUTIONREPORT, {})
    with pytest.raises(
        FIXMessageError,
        match="Missing required field=SchemaField.*OrderID|37",
    ):
        schema.validate(m)

    m = FIXMessage(FMsg.EXECUTIONREPORT, {FTag.OrderID: "1234", "12309812": "12"})
    with pytest.raises(
        FIXMessageError,
        match="msg tag=12309812 not in schema",
    ):
        schema.validate(m)

    m = FIXMessage(FMsg.EXECUTIONREPORT, {FTag.OrderID: "1234", FTag.AdvSide: "1"})
    with pytest.raises(
        FIXMessageError,
        match="msg field=AdvSide.*not allowed in SchemaMessage.*name=ExecutionReport",
    ):
        schema.validate(m)

    m = FIXMessage(
        FMsg.EXECUTIONREPORT, {FTag.OrderID: "1234", FTag.NoContraBrokers: "1"}
    )
    with pytest.raises(
        FIXMessageError,
        match="msg tag=382 val=1 must be a group",
    ):
        schema.validate(m)

    m = FIXMessage(FMsg.EXECUTIONREPORT)
    m.set_group(FTag.OrderID, [{1: "1`231"}])
    with pytest.raises(
        FIXMessageError,
        match="msg tag=37 val=.* must be a tag, got group",
    ):
        schema.validate(m)

    m = FIXMessage(FMsg.EXECUTIONREPORT, {FTag.OrderID: "1234"})
    m.set_group(FTag.NoPartyIDs, [{1: "test"}])
    with (
        patch("asyncfix.protocol.schema.SchemaGroup.validate_group") as mock_valgrp,
        patch("asyncfix.protocol.schema.SchemaField.validate_value") as mock_valfield,
    ):
        schema.validate(m)

    assert mock_valgrp.called
    assert mock_valgrp.call_count == 1
    assert len(mock_valgrp.call_args[0][0]) == 1
    assert mock_valgrp.call_args[1] == {}
    assert mock_valgrp.call_args[0][0][0] == {1: "test"}


def test_field_type_validation__int():
    f = SchemaField("1", "test", "INT")

    assert f.validate_value("1")
    assert f.validate_value("0")
    assert f.validate_value("-1")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("10.2")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("10.0")


def test_field_type_validation__seqnum():
    f = SchemaField("1", "test", "SEQNUM")

    assert f.validate_value("1")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("10.2")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("10.0")

    with pytest.raises(FIXMessageError, match="validation error .*: zero value"):
        assert f.validate_value("0")

    with pytest.raises(FIXMessageError, match="validation error .*: negative value"):
        assert f.validate_value("-1")


def test_field_type_validation__numingroup():
    f = SchemaField("1", "test", "NUMINGROUP")

    assert f.validate_value("1")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("10.2")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("10.0")

    with pytest.raises(FIXMessageError, match="validation error .*: zero value"):
        assert f.validate_value("0")

    with pytest.raises(FIXMessageError, match="validation error .*: negative value"):
        assert f.validate_value("-1")


def test_field_type_validation__dayofmonth():
    f = SchemaField("1", "test", "DAYOFMONTH")

    for i in range(1, 32):
        assert f.validate_value(str(i))

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("10.2")

    with pytest.raises(
        FIXMessageError, match="validation error .*: invalid literal for int"
    ):
        assert f.validate_value("10.0")

    with pytest.raises(FIXMessageError, match="validation error .*: out of range "):
        assert f.validate_value("0")

    with pytest.raises(FIXMessageError, match="validation error .*: out of range"):
        assert f.validate_value("32")


def test_field_type_validation__float():
    f = SchemaField("1", "test", "FLOAT")

    assert f.validate_value("1")
    assert f.validate_value("0.0")
    assert f.validate_value("1.1231")
    assert f.validate_value("-1.12310923810281")

    with pytest.raises(
        FIXMessageError, match="validation error .*: could not convert string to float"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("nan")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("inf")


def test_field_type_validation__qty():
    f = SchemaField("1", "test", "QTY")

    assert f.validate_value("1")
    assert f.validate_value("0.0")
    assert f.validate_value("1.1231")
    assert f.validate_value("-1.12310923810281")

    with pytest.raises(
        FIXMessageError, match="validation error .*: could not convert string to float"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("nan")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("inf")


def test_field_type_validation__price():
    f = SchemaField("1", "test", "PRICE")

    assert f.validate_value("1")
    assert f.validate_value("0.0")
    assert f.validate_value("1.1231")
    assert f.validate_value("-1.12310923810281")

    with pytest.raises(
        FIXMessageError, match="validation error .*: could not convert string to float"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("nan")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("inf")


def test_field_type_validation__priceoffset():
    f = SchemaField("1", "test", "PRICEOFFSET")

    assert f.validate_value("1")
    assert f.validate_value("0.0")
    assert f.validate_value("1.1231")
    assert f.validate_value("-1.12310923810281")

    with pytest.raises(
        FIXMessageError, match="validation error .*: could not convert string to float"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("nan")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("inf")


def test_field_type_validation__amt():
    f = SchemaField("1", "test", "AMT")

    assert f.validate_value("1")
    assert f.validate_value("0.0")
    assert f.validate_value("1.1231")
    assert f.validate_value("-1.12310923810281")

    with pytest.raises(
        FIXMessageError, match="validation error .*: could not convert string to float"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("nan")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("inf")


def test_field_type_validation__percentage():
    f = SchemaField("1", "test", "PERCENTAGE")

    assert f.validate_value("1")
    assert f.validate_value("0.0")
    assert f.validate_value("1.1231")
    assert f.validate_value("-1.12310923810281")

    with pytest.raises(
        FIXMessageError, match="validation error .*: could not convert string to float"
    ):
        assert f.validate_value("as")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("nan")

    with pytest.raises(
        FIXMessageError, match="validation error .*: not isfinite number"
    ):
        assert f.validate_value("inf")


def test_field_type_validation__string():
    f = SchemaField("1", "test", "STRING")

    assert f.validate_value("1")
    assert f.validate_value("A")
    assert f.validate_value("-1.12310923810281")
    assert f.validate_value("Hey this is alphanum string ! Also, some @tags, #test")

    with pytest.raises(
        FIXMessageError, match="validation error .*: Value contains SOH char"
    ):
        assert f.validate_value("a\x01s")

    with pytest.raises(
        FIXMessageError, match="validation error .*: Value contains `=` char"
    ):
        assert f.validate_value("as some tag=values")


def test_field_type_validation__char():
    f = SchemaField("1", "test", "CHAR")

    assert f.validate_value("1")
    assert f.validate_value("A")
    assert f.validate_value("z")
    assert f.validate_value("!")

    with pytest.raises(FIXMessageError, match="max legth exceeded"):
        assert f.validate_value("as")


def test_field_type_validation__boolean():
    f = SchemaField("1", "test", "BOOLEAN")

    assert f.validate_value("Y")
    assert f.validate_value("N")

    with pytest.raises(FIXMessageError, match="out of subset"):
        assert f.validate_value("Z")

    with pytest.raises(FIXMessageError, match="max legth exceeded"):
        assert f.validate_value("ZA")


def test_field_type_validation__multiplestringvalue():
    f = SchemaField("1", "test", "MULTIPLESTRINGVALUE")

    assert f.validate_value("Y AS NA za")
    assert f.validate_value("N N 2 1 n")


def test_field_type_validation__country():
    f = SchemaField("1", "test", "COUNTRY")

    assert f.validate_value("RU")
    assert f.validate_value("US")

    with pytest.raises(FIXMessageError, match="max legth exceeded"):
        assert f.validate_value("ZAR")


def test_field_type_validation__currency():
    f = SchemaField("1", "test", "CURRENCY")

    assert f.validate_value("RUB")
    assert f.validate_value("USD")

    with pytest.raises(FIXMessageError, match="max legth exceeded"):
        assert f.validate_value("EURU")

    with pytest.raises(
        FIXMessageError, match="value contains non alphanumeric letters"
    ):
        assert f.validate_value("EU@")


def test_field_type_validation__exchange():
    f = SchemaField("1", "test", "EXCHANGE")

    assert f.validate_value("NYSE")
    assert f.validate_value("NQ")

    with pytest.raises(FIXMessageError, match="max legth exceeded"):
        assert f.validate_value("EUREx")


def test_field_type_validation__localmktdate():
    f = SchemaField("1", "test", "LOCALMKTDATE")

    assert f.validate_value("20230921")

    with pytest.raises(
        FIXMessageError, match="time data 'EUREx' does not match format '%Y%m%d'"
    ):
        assert f.validate_value("EUREx")

    with pytest.raises(FIXMessageError, match="unconverted data remains"):
        assert f.validate_value("20230921 14:00:00.12312")

    with pytest.raises(FIXMessageError, match="day is out of range for month"):
        assert f.validate_value("20230231")


def test_field_type_validation__utsdateonly():
    f = SchemaField("1", "test", "UTCDATEONLY")

    assert f.validate_value("20230921")

    with pytest.raises(
        FIXMessageError, match="time data 'EUREx' does not match format '%Y%m%d'"
    ):
        assert f.validate_value("EUREx")

    with pytest.raises(FIXMessageError, match="unconverted data remains"):
        assert f.validate_value("20230921 14:00:00.12312")

    with pytest.raises(FIXMessageError, match="day is out of range for month"):
        assert f.validate_value("20230231")


def test_field_type_validation__utctimestamp():
    f = SchemaField("1", "test", "UTCTIMESTAMP")

    assert f.validate_value("20230921-14:00:00")
    assert f.validate_value("20230921-14:00:00.123")
    assert f.validate_value("20230921-14:00:00.123456")

    with pytest.raises(FIXMessageError, match=" does not match format "):
        assert f.validate_value("EUREx")

    with pytest.raises(FIXMessageError, match="unconverted data remains"):
        assert f.validate_value("20230921-14:00:00.123456789")


def test_field_type_validation__utctimeonly():
    f = SchemaField("1", "test", "UTCTIMEONLY")

    assert f.validate_value("14:00:00")
    assert f.validate_value("14:00:00.123")
    assert f.validate_value("14:00:00.123456")

    with pytest.raises(FIXMessageError, match=" does not match format "):
        assert f.validate_value("EUREx")

    with pytest.raises(FIXMessageError, match="unconverted data remains"):
        assert f.validate_value("14:00:00.123456789")

    with pytest.raises(FIXMessageError, match=" does not match format "):
        assert f.validate_value("20230921 14:00:00.123456789")


def test_field_type_validation__monthyear():
    f = SchemaField("1", "test", "MONTHYEAR")

    assert f.validate_value("202309")
    assert f.validate_value("20230921")
    assert f.validate_value("202309w1")
    assert f.validate_value("202309w2")
    assert f.validate_value("202309w3")
    assert f.validate_value("202309w4")
    assert f.validate_value("202309w5")

    with pytest.raises(FIXMessageError, match=" does not match format "):
        assert f.validate_value("EUREx")

    with pytest.raises(FIXMessageError, match="remainder YYYYMM invalid"):
        assert f.validate_value("20230925w5")

    with pytest.raises(FIXMessageError, match="ww must be in w1,w2,w3,w4,w5"):
        assert f.validate_value("202309w6")


def test_field_type_validation__data():
    f = SchemaField("1", "test", "DATA")

    # pass all
    assert f.validate_value("202309\x01")


def test_field_type_validation__length():
    f = SchemaField("1", "test", "LENGTH")

    # pass all
    assert f.validate_value("202309")
    assert f.validate_value("-202309")


def test_field_type_validation__unsupported():
    f = SchemaField("1", "test", "UNSUPPORTED")

    # pass all
    with patch("asyncfix.protocol.schema.warnings.warn") as mock_warn:
        assert f.validate_value("202309")
        assert mock_warn.called
        assert "UNSUPPORTED" in mock_warn.call_args[0][0]


def test_schema_validation_group(fix_simple_xml):
    schema = FIXSchema(fix_simple_xml)
    assert schema[FTag.NoPartyIDs].tag == "453"
    assert schema[453].tag == "453"
    assert schema["453"].tag == "453"
    assert schema["NoPartyIDs"].tag == "453"

    m = schema.messages_types[FMsg.EXECUTIONREPORT]

    assert m.msg_type == FMsg.EXECUTIONREPORT
    g = m["NoPartyIDs"]
    assert isinstance(g, SchemaGroup)

    with pytest.raises(
        FIXMessageError, match="contains unsupported tag for SchemaGroup"
    ):
        msg_g = FIXMessage(
            FMsg.EXECUTIONREPORT,
            {
                FTag.NoPartyIDs: [
                    {FTag.PartyID: "asd", FTag.Account: "1"},
                    {FTag.PartyID: "asd", FTag.PartyRole: "asda"},
                ]
            },
        )
        g.validate_group(msg_g.get_group_list(FTag.NoPartyIDs))

    with pytest.raises(FIXMessageError, match="does not contain mandatory first tag"):
        msg_g = FIXMessage(
            FMsg.EXECUTIONREPORT,
            {
                FTag.NoPartyIDs: [
                    {FTag.PartyRole: "1"},
                    {FTag.PartyID: "asd", FTag.PartyRole: "asda"},
                ]
            },
        )
        g.validate_group(msg_g.get_group_list(FTag.NoPartyIDs))

    g = m["NoContraBrokers"]
    with pytest.raises(
        FIXMessageError, match="missing required field SchemaField.*ContraTrader|"
    ):
        msg_g = FIXMessage(
            FMsg.EXECUTIONREPORT,
            {
                FTag.NoContraBrokers: [
                    {FTag.ContraBroker: "1"},
                ]
            },
        )
        g.validate_group(msg_g.get_group_list(FTag.NoContraBrokers))

    g = m["NoContraBrokers"]
    with pytest.raises(FIXMessageError, match="incorrect tag order"):
        msg_g = FIXMessage(
            FMsg.EXECUTIONREPORT,
            {
                FTag.NoContraBrokers: [
                    {
                        FTag.ContraBroker: "1",
                        FTag.Commission: "10",
                        FTag.ContraTrader: "1",
                    },
                ]
            },
        )
        g.validate_group(msg_g.get_group_list(FTag.NoContraBrokers))

    g = m["NoContraBrokers"]
    with pytest.raises(
        FIXMessageError,
        match="Commission|12 validation error .* could not convert string to float:",
    ):
        msg_g = FIXMessage(
            FMsg.EXECUTIONREPORT,
            {
                FTag.NoContraBrokers: [
                    {
                        FTag.ContraBroker: "1",
                        FTag.ContraTrader: "1",
                        FTag.Commission: "asd",
                    },
                ]
            },
        )
        g.validate_group(msg_g.get_group_list(FTag.NoContraBrokers))


def test_schema_validation_group_nested(fix_simple_xml):
    schema = FIXSchema(fix_simple_xml)
    m = schema.messages_types[FMsg.EXECUTIONREPORT]

    g = m["NoContraBrokers"]

    with pytest.raises(
        FIXMessageError,
        match="tag=382 must be a group ",
    ):
        msg_g = FIXMessage(
            FMsg.EXECUTIONREPORT,
            {
                FTag.NoContraBrokers: [
                    {
                        FTag.ContraBroker: "1",
                        FTag.ContraTrader: "1",
                        FTag.Commission: "10",
                    },
                    {
                        FTag.ContraBroker: "1",
                        FTag.ContraTrader: "1",
                        FTag.Commission: "10",
                        FTag.NoContraBrokers: "2",
                    },
                ]
            },
        )
        g.validate_group(msg_g.get_group_list(FTag.NoContraBrokers))

    with pytest.raises(
        FIXMessageError,
        match=r"fixmessage=\[337=1\] does not contain mandatory first tag ",
    ):
        msg_g = FIXMessage(
            FMsg.EXECUTIONREPORT,
            {
                FTag.NoContraBrokers: [
                    {
                        FTag.ContraBroker: "1",
                        FTag.ContraTrader: "1",
                        FTag.Commission: "10",
                    },
                    {
                        FTag.ContraBroker: "1",
                        FTag.ContraTrader: "1",
                        FTag.Commission: "10",
                        FTag.NoContraBrokers: [{FTag.ContraTrader: "1"}],
                    },
                ]
            },
        )
        g.validate_group(msg_g.get_group_list(FTag.NoContraBrokers))



import unittest

from asyncfix.connection import MessageDirection
from asyncfix.journaler import Journaler
from asyncfix.message import FIXContainer, FIXMessage
from asyncfix.session import FIXSession


class JournalerTests(unittest.TestCase):
    def msg_tag34(self, tag34):
        msg = FIXMessage("AB")
        msg.set("45", "dgd")
        msg.set("32", "aaaa")
        msg.set("323", "bbbb")

        rptgrp1 = FIXContainer()
        rptgrp1.set("611", "aaa")
        rptgrp1.set("612", "bbb")
        rptgrp1.set("613", "ccc")

        msg.add_group("444", rptgrp1, 0)
        msg.set("34", str(tag34))
        return msg

    def testAddExtractMsg(self):
        journal = Journaler()

        session = FIXSession(1, "S1", "T1")

        for i in range(0, 5):
            msg = self.msg_tag34(i)
            journal.persist_msg(msg, session, MessageDirection.OUTBOUND)

        msg = journal.recover_msg(session, MessageDirection.OUTBOUND, 1)

    def testAddExtractMultipleMsgs(self):
        journal = Journaler()

        session = FIXSession(1, "S1", "T1")

        for i in range(0, 5):
            msg = self.msg_tag34(i)
            journal.persist_msg(msg, session, MessageDirection.OUTBOUND)

        msgs = journal.recover_msgs(session, MessageDirection.OUTBOUND, 0, 4)
        for i in range(0, len(msgs)):
            msg = self.msg_tag34(i)
            self.assertEqual(msg, msgs[i])
